# -*- coding: utf-8 -*-
#
# STSCI documentation build configuration file, created by
# sphinx-quickstart on Thu Oct 22 17:25:41 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import sys
from lcviz import __version__

try:
    from sphinx_astropy.conf.v1 import *  # noqa
    from sphinx_astropy.conf.v2 import *  # noqa
except ImportError:
    print('ERROR: the documentation requires the sphinx-astropy package to be installed')
    sys.exit(1)

# -- General configuration ------------------------------------------------

# Configuration for intersphinx
intersphinx_mapping = {
    'jdaviz': ('https://jdaviz.readthedocs.io/en/latest/', None),
    'lightkurve': ('https://lightkurve.github.io/lightkurve/', None)
#     'python': ('https://docs.python.org/3/',
#                (None, 'http://data.astropy.org/intersphinx/python3.inv')),
#     'numpy': ('https://numpy.org/doc/stable/',
#               (None, 'http://data.astropy.org/intersphinx/numpy.inv')),
#     'scipy': ('https://docs.scipy.org/doc/scipy/reference/',
#               (None, 'http://data.astropy.org/intersphinx/scipy.inv')),
#     'matplotlib': ('https://matplotlib.org/',
#                    (None, 'http://data.astropy.org/intersphinx/matplotlib.inv')),  # noqa
#     'astropy': ('https://docs.astropy.org/en/stable/', None)
    }

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx_automodapi.automodapi',
    'sphinx.ext.mathjax',
    'sphinx_design',
    'sphinx_togglebutton']

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project
project = "lcviz"
author = "JDADF Developers"
year = datetime.datetime.now().year
copyright = f'{year}, {author}'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# build documents.
#
# The full version, including alpha/beta/rc tags.
release = __version__
dev = "dev" in release
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'obj'

# Don't show summaries of the members in each class along with the
# class' docstring
numpydoc_show_class_members = False

autosummary_generate = True

automodapi_toctreedirnm = 'api'

# Class documentation should contain *both* the class docstring and
# the __init__ docstring
autoclass_content = "both"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    '-Nfontsize=10',
    '-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Efontsize=10',
    '-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Gfontsize=10',
    '-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif'
]

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.

# html_css_files = ["lcviz.css"]
html_copy_source = False

html_theme_options.update(  # noqa: F405
    {
        "github_url": "https://github.com/spacetelescope/lcviz",
        "use_edit_page_button": True,
    }
)

html_context = {
    "default_mode": "light",
    "to_be_indexed": ["stable", "latest"],
    "is_development": dev,
    "github_user": "spacetelescope",
    "github_repo": "lcviz",
    "github_version": "main",
    "doc_path": "docs",
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'sphinx_rtd_theme'

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'logos/lcviz.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'logos/lcviz.ico'

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['globaltoc.html', 'relations.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = f'{project}doc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '14pt',
    # Additional stuff for the LaTeX preamble.
    'preamble': r'''\usepackage{enumitem} \setlistdepth{99}'''
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  ('index', f'{project}.tex', f'{project} Documentation',
   f'{project}', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
latex_show_urls = 'True'

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = True

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', f'{project}', f'{project} Documentation',
     [f'{project}'], 1)
]

# If true, show URL addresses after external links.
man_show_urls = True

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', f'{project}', f'{project} Documentation',
   f'{author}', f'{project}', f'{project}',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
texinfo_show_urls = 'inline'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = f'{project}'
epub_author = f'{author}'
epub_publisher = f'{author}'
epub_copyright = f'{year} {author}'

# The basename for the epub file. It defaults to the project name.
# epub_basename = u'wfc3tools'

# The HTML theme for the epub output. Since the default themes are not
# optimized for small screen space, using the same theme for HTML and
# epub output is usually not wise. This defaults to 'epub', a theme designed
# to save visual space.
epub_theme = 'epub'

# The language of the text. It defaults to the language option
# or en if the language is not set.
# epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
# epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
# epub_cover = ()

# A sequence of (type, uri, title) tuples for the guide element of content.opf.
# epub_guide = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_post_files = []

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# The depth of the table of contents in toc.ncx.
# epub_tocdepth = 3

# Allow duplicate toc entries.
# epub_tocdup = True

# Choose between 'default' and 'includehidden'.
# epub_tocscope = 'default'

# Fix unsupported image types using the PIL.
# epub_fix_images = False

# Scale large images.
# epub_max_image_width = 0

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# epub_show_urls = 'inline'

# If false, no index is generated.
# epub_use_index = True



def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter
    lc1 = lc
    lc2 = lc.copy()

    lcviz.load_data(lc1, 'lc1')
    lcviz.load_data(lc2, 'lc2')
    lcviz.app.add_data_to_viewer('flux-vs-time', 'lc2')
    # lcviz.show()

    stitch = lcviz.plugins['Stitch']
    stitch.open_in_tray()
    stitch.dataset.select_all()
    stitched_lc = stitch.stitch()
    print(stitched_lc)


def test_plugin_stitch(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter, 'lc1')

    assert "Stitch" not in helper.plugins.keys()

    helper.load_data(light_curve_like_kepler_quarter.copy(), 'lc2')
    helper.app.add_data_to_viewer('flux-vs-time', 'lc2')
    assert "Stitch" in helper.plugins.keys()

    stitch = helper.plugins['Stitch']
    stitch.dataset.select_all()
    stitch.remove_input_datasets = True
    stitched_lc = stitch.stitch()

    assert len(stitched_lc) == 2 * len(light_curve_like_kepler_quarter)
    assert len(helper.app.data_collection) == 1



import numpy as np

from traitlets import Bool, Unicode, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty, IntHandleEmpty
from jdaviz.core.events import ViewerAddedMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin,
                                        AutoTextField,
                                        skip_if_no_updates_since_last_active,
                                        with_spinner, with_temp_disable)
from jdaviz.core.user_api import PluginUserApi

from lcviz.components import FluxColumnSelectMixin
from lcviz.marks import LivePreviewTrend, LivePreviewFlattened
from lcviz.utils import data_not_folded, is_not_tpf
from lcviz.viewers import TimeScatterView, PhaseScatterView
from lcviz.parsers import _data_with_reftime

__all__ = ['Flatten']


@tray_registry('flatten', label="Flatten")
class Flatten(PluginTemplateMixin, FluxColumnSelectMixin, DatasetSelectMixin):
    """
    See the :ref:`Flatten Plugin Documentation <flatten>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``show_live_preview`` : bool
        Whether to show the live-preview of the (unnormalized) flattened light curve
    * ``show_trend_preview`` : bool
        Whether to show the live-preview of the trend curve used to flatten the light curve
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to flatten.
    * ``window_length`` : int
    * ``polyorder`` : int
    * ``break_tolerance`` : int
    * ``niters`` : int
    * ``sigma`` : float
    * ``unnormalize`` : bool
    * ``flux_label`` (:class:`~jdaviz.core.template_mixin.AutoTextField`):
      Label for the resulting flux column added to ``dataset`` and automatically selected as the new
      flux column (origin).
    * :meth:`flatten`
    """
    template_file = __file__, "flatten.vue"
    uses_active_status = Bool(True).tag(sync=True)

    show_live_preview = Bool(True).tag(sync=True)
    show_trend_preview = Bool(True).tag(sync=True)
    flatten_err = Unicode().tag(sync=True)

    window_length = IntHandleEmpty(101).tag(sync=True)
    polyorder = IntHandleEmpty(2).tag(sync=True)
    break_tolerance = IntHandleEmpty(5).tag(sync=True)
    niters = IntHandleEmpty(3).tag(sync=True)
    sigma = FloatHandleEmpty(3).tag(sync=True)
    unnormalize = Bool(False).tag(sync=True)

    flux_label_label = Unicode().tag(sync=True)
    flux_label_default = Unicode().tag(sync=True)
    flux_label_auto = Bool(True).tag(sync=True)
    flux_label_invalid_msg = Unicode('').tag(sync=True)
    flux_label_overwrite = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._plugin_description = 'Flatten input light curve.'
        self.flux_label = AutoTextField(self, 'flux_label_label',
                                        'flux_label_default', 'flux_label_auto',
                                        'flux_label_invalid_msg')

        # do not support flattening data in phase-space
        # do not allow TPF as input
        self.dataset.add_filter(data_not_folded, is_not_tpf)

        # marks do not exist for the new viewer, so force another update to compute and draw
        # those marks
        self.hub.subscribe(self, ViewerAddedMessage, handler=lambda _: self._live_update())

        self._set_default_label()

    @property
    def user_api(self):
        expose = ['show_live_preview', 'show_trend_preview',
                  'dataset',
                  'window_length', 'polyorder', 'break_tolerance',
                  'niters', 'sigma', 'unnormalize', 'flux_label', 'flatten']
        return PluginUserApi(self, expose=expose)

    @property
    def marks(self):
        trend_marks = {}
        flattened_marks = {}

        for viewer in self.app._viewer_store.values():
            needs_trend = isinstance(viewer, TimeScatterView) and not isinstance(viewer, PhaseScatterView)  # noqa
            needs_flattened = isinstance(viewer, (TimeScatterView, PhaseScatterView))
            for mark in viewer.figure.marks:
                if isinstance(mark, LivePreviewTrend):
                    trend_marks[viewer.reference] = mark
                    needs_trend = False
                elif isinstance(mark, LivePreviewFlattened):
                    flattened_marks[viewer.reference] = mark
                    needs_flattened = False
                if not needs_trend and not needs_flattened:
                    break
            if needs_trend:
                mark = LivePreviewTrend(viewer, visible=self.is_active)
                viewer.figure.marks = viewer.figure.marks + [mark]
                trend_marks[viewer.reference] = mark
            if needs_flattened:
                mark = LivePreviewFlattened(viewer, visible=self.is_active)
                viewer.figure.marks = viewer.figure.marks + [mark]
                flattened_marks[viewer.reference] = mark

        return trend_marks, flattened_marks

    @observe('dataset_selected', 'flux_column_selected')
    def _set_default_label(self, event={}):
        '''Generate a label and set the results field to that value'''
        if not hasattr(self, 'dataset'):  # pragma: no cover
            return

        # TODO: have an option to create new data entry and drop other columns?
        # (or should that just go through future data cloning)
        self.flux_label.default = f"{self.flux_column_selected}_flattened"

    @observe('flux_label_label', 'dataset')
    def _update_label_valid(self, event={}):
        if self.flux_label.value in self.flux_column.choices:
            self.flux_label.invalid_msg = ''
            self.flux_label_overwrite = True
        elif self.flux_label.value in getattr(self.dataset.selected_obj, 'columns', []):
            self.flux_label.invalid_msg = 'name already in use'
        else:
            self.flux_label.invalid_msg = ''
            self.flux_label_overwrite = False

    @with_spinner()
    def flatten(self, add_data=True):
        """
        Flatten the input light curve (``dataset``) using lightkurve.flatten.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting light curve as a flux column and select that as the new
            flux column (origin) for that data entry.

        Returns
        -------
        output_lc : `~lightkurve.LightCurve`
            The flattened light curve.
        trend_lc : `~lightkurve.LightCurve`
            The trend used to flatten the light curve.
        """
        input_lc = self.dataset.selected_obj
        if input_lc is None:  # pragma: no cover
            raise ValueError("no input dataset selected")

        output_lc, trend_lc = input_lc.flatten(return_trend=True,
                                               window_length=self.window_length,
                                               polyorder=self.polyorder,
                                               break_tolerance=self.break_tolerance,
                                               niters=self.niters,
                                               sigma=self.sigma)

        if self.unnormalize:
            factor = np.nanmedian(trend_lc.flux.value)
            output_lc.flux *= factor
            output_lc.flux_err *= factor
            output_lc.meta['NORMALIZED'] = False

        if add_data:
            # add data as a new flux and corresponding err columns in the existing data entry
            # and select as flux column (origin)
            data = _data_with_reftime(self.app, output_lc)
            self.flux_column.add_new_flux_column(flux=data['flux'],
                                                 flux_err=data['flux_err'],
                                                 label=self.flux_label.value,
                                                 selected=True)

        return output_lc, trend_lc

    def _clear_marks(self):
        for mark_set in self.marks:
            for mark in mark_set.values():
                if mark.visible:
                    mark.clear()
                    mark.visible = False

    @observe('is_active', 'show_live_preview', 'show_trend_preview')
    def _toggle_marks(self, event={}):
        live_visible = self.show_live_preview and self.is_active
        trend_visible = self.show_trend_preview and self.is_active

        trend_marks, flattened_marks = self.marks
        for mark in trend_marks.values():
            mark.visible = trend_visible
        for mark in flattened_marks.values():
            mark.visible = live_visible

        if ((live_visible or trend_visible) and
                event.get('name') in ('is_active', 'show_live_preview', 'show_trend_preview')):
            # then the marks themselves need to be updated
            self._live_update(event)

    @observe('dataset_selected', 'flux_column_selected',
             'window_length', 'polyorder', 'break_tolerance',
             'niters', 'sigma', 'previews_temp_disabled')
    @skip_if_no_updates_since_last_active()
    @with_temp_disable(0.3)
    def _live_update(self, event={}):
        if self.dataset_selected == '' or self.flux_column_selected == '':
            self._clear_marks()
            return

        try:
            output_lc, trend_lc = self.flatten(add_data=False)
        except Exception as e:
            self.flatten_err = str(e)
            self._clear_marks()
            return
        self.flatten_err = ''

        if event.get('name') not in ('is_active', 'show_live_preview', 'show_trend_preview'):
            # mark visibility hasn't been handled yet
            self._toggle_marks(event)

        if self.unnormalize:
            output_flux = output_lc.flux.value
        else:
            output_flux = output_lc.flux.value * np.nanmedian(trend_lc.flux.value)

        ref_time = trend_lc.meta.get('reference_time', 0)
        times = trend_lc.time - ref_time
        trend_marks, flattened_marks = self.marks
        for mark in trend_marks.values():
            # TODO: need to account for phasing
            mark.update_ty(times.value, trend_lc.flux.value)
        for mark in flattened_marks.values():
            mark.update_ty(times.value, output_flux)

    def vue_apply(self, *args, **kwargs):
        try:
            self.flatten(add_data=True)
        except Exception as e:
            self.flatten_err = str(e)
        else:
            self.flatten_err = ''



import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time
import astropy.units as u
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive

from traitlets import Bool, Float, List, Unicode, observe

from glue.core.link_helpers import LinkSame
from glue.core.message import DataCollectionAddMessage
from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import (NewViewerMessage, ViewerAddedMessage, ViewerRemovedMessage)
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, DatasetSelectMixin,
                                        SelectPluginComponent, EditableSelectPluginComponent,
                                        with_spinner)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.events import SnackbarMessage

from lightkurve import periodogram, FoldedLightCurve

from lcviz.events import EphemerisComponentChangedMessage, EphemerisChangedMessage
from lcviz.viewers import PhaseScatterView
from lcviz.utils import is_not_tpf

__all__ = ['Ephemeris']

_default_t0 = 0.0
_default_period = 1.0
_default_dpdt = 0.0
_default_wrap_at = 1.0

_default_query_radius = 2  # [arcsec]


@tray_registry('ephemeris', label="Ephemeris")
class Ephemeris(PluginTemplateMixin, DatasetSelectMixin):
    """
    See the :ref:`Ephemeris Plugin Documentation <ephemeris>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``component`` (:class:`~jdaviz.core.template_mixin.EditableSelectPluginComponent`):
      Label of the component corresponding to the active ephemeris.
    * :attr:`t0` : float
      Zeropoint of the ephemeris.
    * :attr:`period` : float
      Period of the ephemeris, defined at ``t0``.
    * :attr:`dpdt` : float
      First derivative of the period of the ephemeris.
    * :attr:`wrap_at` : float
      Phase at which to wrap (phased data will encompass the range 1-wrap_at to wrap_at).
    * :meth:`ephemeris`
    * :meth:`ephemerides`
    * :meth:`update_ephemeris`
    * :meth:`create_phase_viewer`
    * :meth:`add_component`
    * :meth:`rename_component`
    * :meth:`times_to_phases`
    * :meth:`phases_to_times`
    * :meth:`get_data`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to use for determining the period.
    * ``method`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Method/algorithm to determine the period.
    * :meth:`query_for_ephemeris`
      Query the `NASA Exoplanet Archive <https://exoplanetarchive.ipac.caltech.edu/>`_'s
      `Planetary System Composite Parameters
      <https://exoplanetarchive.ipac.caltech.edu/docs/pscp_about.html>`_
      table for the planet-hosting star identified
      by the observation's header key "OBJECT", or if that fails,
      by the observation's header keys for RA and Dec.
    * ``query_result`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      The name of a planet from a NASA Exoplanet Archive query, used for
      adopting literature values for the orbital period and mid-transit time.
    * :meth:`create_ephemeris_from_query`
      Create an ephemeris component with the period and epoch from
      the planet selected from the NASA Exoplanet Archive query in ``query_result``.
    """
    template_file = __file__, "ephemeris.vue"

    # EPHEMERIS
    component_mode = Unicode().tag(sync=True)
    component_edit_value = Unicode().tag(sync=True)
    component_items = List().tag(sync=True)
    component_selected = Unicode().tag(sync=True)

    phase_viewer_exists = Bool(False).tag(sync=True)

    t0 = FloatHandleEmpty(_default_t0).tag(sync=True)
    t0_step = Float(0.1).tag(sync=True)
    period = FloatHandleEmpty(_default_period).tag(sync=True)
    period_step = Float(0.1).tag(sync=True)
    dpdt = FloatHandleEmpty(_default_dpdt).tag(sync=True)
    dpdt_step = Float(0.1).tag(sync=True)
    wrap_at = FloatHandleEmpty(_default_wrap_at).tag(sync=True)

    # PERIOD FINDING
    method_items = List().tag(sync=True)
    method_selected = Unicode().tag(sync=True)

    method_spinner = Bool().tag(sync=True)
    method_err = Unicode().tag(sync=True)

    period_at_max_power = Float().tag(sync=True)

    # QUERIES
    query_name = Unicode().tag(sync=True)
    query_ra = FloatHandleEmpty().tag(sync=True)
    query_dec = FloatHandleEmpty().tag(sync=True)
    query_radius = FloatHandleEmpty(_default_query_radius).tag(sync=True)
    query_result_items = List().tag(sync=True)
    query_result_selected = Unicode().tag(sync=True)
    ra_dec_step = Float(0.01).tag(sync=True)
    period_from_catalog = FloatHandleEmpty().tag(sync=True)
    t0_from_catalog = FloatHandleEmpty().tag(sync=True)
    query_spinner = Bool().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._plugin_description = 'Ephemerides for phase-folding.'

        self._default_initialized = False
        self._ignore_ephem_change = False
        self._ephemerides = {}
        self._prev_wrap_at = _default_wrap_at
        self._nasa_exoplanet_archive = None

        self.dataset.add_filter(is_not_tpf)

        self.component = EditableSelectPluginComponent(self,
                                                       name='ephemeris',
                                                       mode='component_mode',
                                                       edit_value='component_edit_value',
                                                       items='component_items',
                                                       selected='component_selected',
                                                       manual_options=['default'],
                                                       on_add=self._on_component_add,
                                                       on_rename_after_selection=self._on_component_rename,  # noqa
                                                       on_remove_after_selection=self._on_component_remove,  # noqa
                                                       validate_choice=self._validate_component)

        # force the original entry in ephemerides with defaults
        self._change_component()

        self.method = SelectPluginComponent(self,
                                            items='method_items',
                                            selected='method_selected',
                                            manual_options=['Lomb-Scargle', 'Box Least Squares'])

        self.query_result = SelectPluginComponent(self,
                                                  items='query_result_items',
                                                  selected='query_result_selected')

        # TODO: could optimize by only updating for the new data entry only
        # (would require some refactoring and probably wouldn't have significant gains)
        self.hub.subscribe(self, DataCollectionAddMessage, handler=self._update_all_phase_arrays)
        self.hub.subscribe(self, ViewerAddedMessage, handler=self._check_if_phase_viewer_exists)
        self.hub.subscribe(self, ViewerRemovedMessage, handler=self._check_if_phase_viewer_exists)

    @property
    def user_api(self):
        expose = [
            'component', 'period', 'dpdt', 't0', 'wrap_at',
            'ephemeris', 'ephemerides',
            'update_ephemeris', 'create_phase_viewer',
            'add_component', 'remove_component', 'rename_component',
            'times_to_phases', 'phases_to_times', 'get_data',
            'dataset', 'method', 'period_at_max_power',
            'adopt_period_at_max_power',
            'query_name', 'query_ra', 'query_dec', 'query_radius',
            'query_for_ephemeris', 'query_result', 'create_ephemeris_from_query'
        ]
        return PluginUserApi(self, expose=expose)

    def _phase_comp_lbl(self, component=None):
        if component is None:
            component = self.component_selected
        if self.app._jdaviz_helper is None:
            # duplicate logic from helper in case this is ever called before the helper
            # is fully intialized
            return f'phase:{component}'
        return self.app._jdaviz_helper._phase_comp_lbl(component)

    @property
    def phase_comp_lbl(self):
        return self._phase_comp_lbl()

    def _generate_phase_viewer_id(self, component=None):
        if component is None:
            component = self.component_selected
        return self.app._jdaviz_helper._get_clone_viewer_reference(f'flux-vs-phase:{component}')

    def _get_phase_viewers(self, lbl=None):
        if lbl is None:
            lbl = self.component_selected
        return [viewer for vid, viewer in self.app._viewer_store.items()
                if isinstance(viewer, PhaseScatterView)
                and viewer._ephemeris_component == lbl]

    @property
    def default_phase_viewer(self):
        if not self.phase_viewer_exists:
            return None
        # we'll just treat the "default" as the first viewer connected to this
        # ephemeris component
        return self._get_phase_viewers()[0]

    @property
    def ephemerides(self):
        return self._ephemerides

    @property
    def ephemeris(self):
        return self.ephemerides.get(self.component_selected, {})

    def _times_to_phases_callable(self, component):
        if component == self.component_selected:
            # retrieving from traitlets is cheaper than dictionaries
            t0 = self.t0
            period = self.period
            dpdt = self.dpdt
            wrap_at = self.wrap_at
        else:
            ephem = self.ephemerides.get(component, {})
            t0 = ephem.get('t0', _default_t0)
            period = ephem.get('period', _default_period)
            dpdt = ephem.get('dpdt', _default_dpdt)
            wrap_at = ephem.get('wrap_at', _default_wrap_at)

        def _callable(times):
            if hasattr(times, '__len__') and not len(times):
                return []
            if dpdt != 0:
                return np.mod(1./dpdt * np.log(1 + dpdt/period*(times-t0)) + (1-wrap_at), 1.0) - (1-wrap_at)  # noqa
            else:
                return np.mod((times-t0)/period + (1-wrap_at), 1.0) - (1-wrap_at)

        return _callable

    def times_to_phases(self, times, ephem_component=None):
        if ephem_component is None:
            ephem_component = self.component.selected

        return self._times_to_phases_callable(ephem_component)(times)

    def phases_to_times(self, phases, ephem_component=None):
        if ephem_component is None:
            ephem_component = self.component.selected

        # this is not used internally, so we don't need the traitlet
        # and callable optimizations
        ephem = self.ephemerides.get(ephem_component, {})
        t0 = ephem.get('t0', _default_t0)
        period = ephem.get('period', _default_period)
        dpdt = ephem.get('dpdt', _default_dpdt)

        if dpdt != 0:
            return t0 + period/dpdt*(np.exp(dpdt*(phases))-1.0)
        else:
            return t0 + (phases)*period

    def _update_all_phase_arrays(self, *args, ephem_component=None):
        # `ephem_component` is the name given to the
        # *ephemeris* component in the orbiting system, e.g. "default",
        # rather than the glue Data Component ID:

        if ephem_component is None:
            for ephem_component in self.component.choices:
                self._update_all_phase_arrays(ephem_component=ephem_component)
            return

        dc = self.app.data_collection

        phase_comp_lbl = self._phase_comp_lbl(ephem_component)

        # we'll create the callable function for this component once so it can be re-used
        _times_to_phases = self._times_to_phases_callable(ephem_component)

        new_links = []
        for i, data in enumerate(dc):
            data_is_folded = '_LCVIZ_EPHEMERIS' in data.meta.keys()
            if data_is_folded:
                continue

            times = data.get_component('World 0').data
            phases = _times_to_phases(times)

            self.app._jdaviz_helper._set_data_component(
                data, phase_comp_lbl, phases
            )

            if i != 0:
                ref_data = dc[0]
                new_link = LinkSame(
                    cid1=ref_data.world_component_ids[0],
                    cid2=data.world_component_ids[0],
                    data1=ref_data,
                    data2=data,
                    labels1=ref_data.label,
                    labels2=data.label
                )

                new_links.append(new_link)

        dc.add_link(new_links)

        # update any plugin markers
        for viewer in self._get_phase_viewers(ephem_component):
            for mark in viewer.custom_marks:
                if hasattr(mark, 'update_phase_folding'):
                    mark.update_phase_folding()

        return phase_comp_lbl

    def create_phase_viewer(self, ephem_component=None):
        """
        Create a new phase viewer corresponding to ``component`` and populate the phase arrays
        with the current ephemeris, if necessary.

        Parameters
        ----------
        ephem_component : str, optional
            label of the component.  If not provided or ``None``, will default to plugin value.
        """
        if ephem_component is None:
            ephem_component = self.component_selected
        phase_comp_lbl = self._phase_comp_lbl(ephem_component)
        dc = self.app.data_collection

        # check to see if this component already has a phase array.  We'll just check the first
        # item in the data-collection since the rest of the logic in this plugin /should/ populate
        # the arrays across all entries.
        if phase_comp_lbl not in [comp.label for comp in dc[0].components]:
            self.update_ephemeris()  # calls _update_all_phase_arrays

        phase_viewer_id = self._generate_phase_viewer_id(ephem_component)
        # TODO: stack horizontally by default?
        self.app._on_new_viewer(NewViewerMessage(PhaseScatterView, data=None, sender=self.app),
                                vid=phase_viewer_id, name=phase_viewer_id,
                                open_data_menu_if_empty=False)

        # access new viewer, set bookkeeping for ephemeris component
        pv = self.app.get_viewer(phase_viewer_id)
        pv._ephemeris_component = ephem_component
        # since we couldn't set ephemeris_component right away, _check_if_phase_viewer_exists
        # might be out-of-date
        self._check_if_phase_viewer_exists()

        # set default data visibility
        tvdm = self.app._jdaviz_helper.default_time_viewer.data_menu
        visible_layers = tvdm.data_labels_visible
        loaded_layers = pv.data_menu.data_labels_loaded
        for data in dc:
            if data.ndim > 1:
                # skip image/cube entries
                continue
            visible = data.label in visible_layers
            if data.label not in loaded_layers:
                pv.data_menu.add_data(data.label)
            pv.data_menu.set_layer_visibility(data.label, visible)

        # set x_att
        phase_comp = self.app._jdaviz_helper._component_ids[phase_comp_lbl]
        pv.state.x_att = phase_comp

        # set viewer limits
        pv.state.x_min, pv.state.x_max = (self.wrap_at-1, self.wrap_at)

        return pv.user_api

    def vue_create_phase_viewer(self, *args):
        if not self.phase_viewer_exists:
            self.create_phase_viewer()

    def vue_period_halve(self, *args):
        self.period /= 2

    def vue_period_double(self, *args):
        self.period *= 2

    def _check_if_phase_viewer_exists(self, *args):
        self.phase_viewer_exists = len(self._get_phase_viewers()) > 0

    def _validate_component(self, lbl):
        if '[' in lbl or ']' in lbl:
            return 'cannot contain square brackets'
        if ':' in lbl:
            return 'cannot contain colon'
        return ''

    def _on_component_add(self, lbl):
        self.hub.broadcast(EphemerisComponentChangedMessage(old_lbl=None, new_lbl=lbl,
                                                            sender=self))

    def _on_component_rename(self, old_lbl, new_lbl):
        # this is triggered when the plugin component detects a change to the component name
        self._ephemerides[new_lbl] = self._ephemerides.pop(old_lbl, {})
        for viewer in self._get_phase_viewers(old_lbl):
            self.app._update_viewer_reference_name(
                viewer._ref_or_id,
                viewer._ref_or_id.replace(old_lbl, new_lbl),
                update_id=True
            )
            viewer._ephemeris_component = new_lbl

        # update metadata entries so that they can be used for filtering applicable entries in
        # data menus
        for dc_item in self.app.data_collection:
            if dc_item.meta.get('_LCVIZ_EPHEMERIS', {}).get('ephemeris', None) == old_lbl:
                dc_item.meta['_LCVIZ_EPHEMERIS']['ephemeris'] = new_lbl
        for data_item in self.app.state.data_items:
            if data_item.get('meta', {}).get('_LCVIZ_EPHEMERIS', {}).get('ephemeris', None) == old_lbl:  # noqa
                data_item['meta']['_LCVIZ_EPHEMERIS']['ephemeris'] = new_lbl

        self._check_if_phase_viewer_exists()
        self.hub.broadcast(EphemerisComponentChangedMessage(old_lbl=old_lbl, new_lbl=new_lbl,
                                                            sender=self))

    def _on_component_remove(self, lbl):
        _ = self._ephemerides.pop(lbl, {})
        # remove the corresponding viewer(s), if any exist
        for viewer in self._get_phase_viewers(lbl):
            self.app.vue_destroy_viewer_item(viewer._ref_or_id)
        self.hub.broadcast(EphemerisComponentChangedMessage(old_lbl=lbl, new_lbl=None,
                                                            sender=self))

    def rename_component(self, old_lbl, new_lbl):
        # NOTE: the component will call _on_component_rename after updating
        self.component.rename_choice(old_lbl, new_lbl)

    def add_component(self, lbl, set_as_selected=True):
        self.component.add_choice(lbl, set_as_selected=set_as_selected)

    def remove_component(self, lbl):
        # NOTE: the component will call _on_component_remove after updating
        self.component.remove_choice(lbl)

    @observe('component_selected')
    def _change_component(self, *args):
        if not hasattr(self, 'component'):
            # plugin/traitlet startup
            return
        if self.component_selected == '':
            # no component selected (this can happen when removing all components)
            return
        self._check_if_phase_viewer_exists()
        ephem = self._ephemerides.get(self.component_selected, {})

        # we'll temporarily disable updating the phasing so that we can set all
        # traitlets simultaneously and THEN revising phase arrays if necessary
        self._ignore_ephem_change = True
        self.t0 = ephem.get('t0', self.t0)
        self.period = ephem.get('period', self.period)
        self.dpdt = ephem.get('dpdt', self.dpdt)
        self.wrap_at = ephem.get('wrap_at', self.wrap_at)

        # if this is a new component, update those default values back to the dictionary
        self.update_ephemeris(t0=self.t0, period=self.period, dpdt=self.dpdt, wrap_at=self.wrap_at)
        self._ignore_ephem_change = False
        if ephem:
            # if there were any changes applied by accessing the dictionary,
            # then we need to update phasing, etc (since we set _ignore_ephem_change)
            # otherwise, this is a new component and there is no need.
            self._ephem_traitlet_changed()

    def update_ephemeris(self, ephem_component=None, t0=None, period=None, dpdt=None, wrap_at=None):
        """
        Update the ephemeris for a given component.

        Parameters
        ----------
        ephem_component : str, optional
            label of the component.  If not provided or ``None``, will default to plugin value.
        t0 : float, optional
            value of t0 to replace
        period : float, optional
            value of period to replace
        dpdt : float, optional
            value of dpdt to replace
        wrap_at : float, optional
            value of wrap_at to replace

        Returns
        -------
        dictionary of ephemeris corresponding to ``component``
        """
        if ephem_component is None:
            ephem_component = self.component_selected

        if ephem_component not in self.component.choices:  # pragma: no cover
            raise ValueError(f"component must be one of {self.component.choices}")

        existing_ephem = self._ephemerides.get(ephem_component, {})
        for name, value in {'t0': t0, 'period': period, 'dpdt': dpdt, 'wrap_at': wrap_at}.items():
            if value is not None:
                existing_ephem[name] = value
                if ephem_component == self.component_selected:
                    setattr(self, name, value)

        self._ephemerides[ephem_component] = existing_ephem
        self._update_all_phase_arrays(ephem_component=ephem_component)
        self.hub.broadcast(EphemerisChangedMessage(ephemeris_label=ephem_component,
                                                   sender=self))
        return existing_ephem

    @observe('period', 'dpdt', 't0', 'wrap_at')
    def _ephem_traitlet_changed(self, event={}):
        if self._ignore_ephem_change:
            return
        for value in (self.period, self.dpdt, self.t0, self.wrap_at):
            if not isinstance(value, (int, float)):
                return
        if self.period <= 0:
            return

        def round_to_1(x):
            return round(x, -int(np.floor(np.log10(abs(x)))))

        # if phase-viewer doesn't yet exist in the app, create it now
        if not self.phase_viewer_exists:
            self.create_phase_viewer()

        # update value in the dictionary (to support multi-ephems)
        if event:
            self.update_ephemeris(**{event.get('name'): event.get('new')})
            # will call _update_all_phase_arrays
        else:
            self._update_all_phase_arrays(ephem_component=self.component_selected)

        # update zoom-limits if wrap_at was changed
        if event.get('name') == 'wrap_at':
            old = event.get('old') if event.get('old') != '' else self._prev_wrap_at
            if event.get('new') != '':
                delta_phase = event.get('new') - old
                for pv in self._get_phase_viewers():
                    pvs = pv.state
                    pvs.x_min, pvs.x_max = pvs.x_min + delta_phase, pvs.x_max + delta_phase
                # we need to cache the old value since it could become a string
                # if the widget is cleared
                self._prev_wrap_at = event.get('new')

        # update step-sizes
        self.period_step = round_to_1(self.period/5000)
        self.dpdt_step = max(round_to_1(abs(self.period * self.dpdt)/1000) if self.dpdt != 0 else 0,
                             1./1000000)
        self.t0_step = round_to_1(self.period/1000)

        if not self._default_initialized:
            # other plugins that use EphemerisSelect don't see the first entry yet
            self._default_initialized = True
            self._on_component_add(self.component_selected)

    @observe('dataset_selected', 'method_selected')
    def _update_periodogram(self, *args):
        if not (hasattr(self, 'method') and hasattr(self, 'dataset')):
            return
        # TODO: support multiselect on self.dataset and combine light curves (or would that be a
        # dedicated plugin of its own)?
        self.method_spinner = True
        self.method_err = ''
        if self.method == 'Box Least Squares':
            try:
                per = periodogram.BoxLeastSquaresPeriodogram.from_lightcurve(self.dataset.selected_obj)  # noqa
            except Exception as err:
                self.method_spinner = False
                self.method_err = str(err)
                return
        elif self.method == 'Lomb-Scargle':
            try:
                per = periodogram.LombScarglePeriodogram.from_lightcurve(self.dataset.selected_obj)
            except Exception as err:
                self.method_spinner = False
                self.method_err = str(err)
                return
        else:  # pragma: no cover
            self.method_spinner = False
            raise NotImplementedError(f"periodogram not implemented for {self.method}")

        # TODO: will need to return in display units once supported
        self.period_at_max_power = per.period_at_max_power.value
        self.method_spinner = False

    def adopt_period_at_max_power(self):
        self.period = self.period_at_max_power

    def vue_adopt_period_at_max_power(self, *args):
        self.adopt_period_at_max_power()

    def get_data(self, dataset, ephem_component=None):
        # TODO: support subset_to_apply and then include a wrapper at the helper-level?
        # (would need to catch when cls does not result in a lightkurve object or write
        # behaviors for other cases as well)
        if ephem_component is None:
            ephem_component = self.component.selected

        lc = self.app._jdaviz_helper.get_data(dataset)
        data = next((x for x in self.app.data_collection if x.label == dataset))

        comps = {str(comp): comp for comp in data.components}
        xcomp = f'phase:{ephem_component}'
        phases = data.get_component(comps.get(xcomp)).data

        # the following code is adopted directly from lightkurve
        # 2. Create the folded object
        phlc = FoldedLightCurve(data=lc)
        # 3. Restore the folded time
        with phlc._delay_required_column_checks():
            phlc.remove_column("time")
            # TODO: phased lc shouldn't have the same time format/scale, but this is needed
            # in order for binning to work (until there's a fix to lightkurve)
            phlc.add_column(Time(phases, format=lc.time.format, scale=lc.time.scale),
                            name="time", index=0)
            phlc.add_column(lc.time.copy(), name="time_original", index=len(lc._required_columns))

        # Add extra column and meta data specific to FoldedLightCurve
        ephemeris = self.ephemerides.get(ephem_component)
        phlc.meta["_LCVIZ_EPHEMERIS"] = {'ephemeris': ephem_component, **ephemeris}
        phlc.meta["PERIOD"] = ephemeris.get('period')
        phlc.meta["EPOCH_TIME"] = ephemeris.get('t0')
        phlc.sort("time")

        return phlc

    @property
    def nasa_exoplanet_archive(self):
        if self._nasa_exoplanet_archive is None:
            self._nasa_exoplanet_archive = NasaExoplanetArchive()

        return self._nasa_exoplanet_archive

    @observe('dataset_selected')
    def _query_params_from_metadata(self, *args):
        self.query_name = self.dataset.selected_obj.meta.get('OBJECT', '')
        self.query_ra = self.dataset.selected_obj.meta.get('RA')
        self.query_dec = self.dataset.selected_obj.meta.get('DEC')

    def query_for_ephemeris(self):
        query_result = None

        if self.query_name:
            # first query by object name:
            query_result = self.nasa_exoplanet_archive.query_object(
                object_name=self.query_name,
                table='pscomppars'
            )

        if (
            (query_result is None or len(query_result) == 0) and
            (None not in (self.query_ra, self.query_dec))
        ):
            # next query by coordinates:
            coord = SkyCoord(ra=self.query_ra, dec=self.query_dec, unit=u.deg)
            query_result = self.nasa_exoplanet_archive.query_region(
                table='pscomppars',
                coordinates=coord,
                radius=self.query_radius * u.arcsec,
            )

        if query_result is None or len(query_result) == 0:
            # no metadata found for RA, Dec, or object name
            return None

        else:
            query_result.sort('pl_name')
            self.astroquery_result = query_result
            self.astroquery_result.add_index('pl_name')
            self.query_result_items = [
                {
                    'label': name,  # required key for SelectPluginComponent
                    'period': period,
                    'epoch': epoch if not np.isnan(epoch) else 0
                }
                for name, period, epoch in zip(
                    list(self.astroquery_result['pl_name']),
                    np.array(self.astroquery_result['pl_orbper'].to_value(u.day)),
                    np.array(self.astroquery_result['pl_tranmid'].to_value(u.day))
                )
            ]

    @observe('query_result_selected')
    def _select_query_result(self, *args):
        selected_query_result = self.astroquery_result.loc[self.query_result_selected]
        self.period_from_catalog = selected_query_result['pl_orbper'].base.to_value(u.day)
        ref_time = self.app.data_collection[0].coords.reference_time.jd
        if np.isnan(selected_query_result['pl_tranmid'].base.to_value(u.day)):
            self.t0_from_catalog = 0
        else:
            self.t0_from_catalog = (
                selected_query_result['pl_tranmid'].base.to_value(u.day) - ref_time
            ) % self.period_from_catalog

    @with_spinner('query_spinner')
    def vue_query_for_ephemeris(self, *args):
        self.query_for_ephemeris()

    def create_ephemeris_from_query(self, *args):
        new_component_label = self.query_result_selected.replace(' ', '')
        if new_component_label in self.component.choices:
            # warn the user that an ephemeris component already exists with this label,
            # a second won't be added:
            self.hub.broadcast(
                SnackbarMessage(
                    f"Ephemeris component {new_component_label} already exists, "
                    f"this ephemeris component will not be added.",
                    sender=self, color="warning"
                )
            )
        elif not np.any(np.isnan([self.period_from_catalog, self.t0_from_catalog])):
            self.add_component(new_component_label)
            self.create_phase_viewer()

            self.period = self.period_from_catalog
            self.t0 = self.t0_from_catalog

            # reset the phase axis wrap to feature the primary transit:
            self.wrap_at = 0.5
            viewer = self._get_phase_viewers()[0]
            viewer.reset_limits()
        else:
            self.hub.broadcast(
                SnackbarMessage(
                    f"Catalog period ({self.period_from_catalog}) or "
                    f"epoch ({self.t0_from_catalog}) is NaN, this ephemeris "
                    f"component will not be added.",
                    sender=self, color="warning"
                )
            )

    def vue_create_ephemeris_from_query(self, *args):
        self.create_ephemeris_from_query()



import os

from glue.config import viewer_tool
from glue.viewers.common.tool import Tool

from jdaviz.core.tools import SidebarShortcutPlotOptions, SidebarShortcutExportPlot

ICON_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), 'data', 'icons'))

# point to the lcviz-version of plot options instead of jdaviz's
SidebarShortcutPlotOptions.plugin_name = 'lcviz-plot-options'
SidebarShortcutExportPlot.plugin_name = 'lcviz-export'


__all__ = ['ViewerClone']


@viewer_tool
class ViewerClone(Tool):
    icon = os.path.join(ICON_DIR, 'viewer_clone')
    tool_id = 'lcviz:viewer_clone'
    action_text = 'Clone viewer'
    tool_tip = 'Clone this viewer'

    def activate(self):
        self.viewer.clone_viewer()



from astropy.io.fits import getheader
import astropy.units as u
import ipyvue
import os

from lightkurve import LightCurve

from glue.config import settings as glue_settings
from glue.core.component_id import ComponentID
from glue.core.link_helpers import LinkSame
from glue.core.units import unit_converter
from jdaviz.core.helpers import ConfigHelper

from lcviz import __version__
from lcviz.viewers import TimeScatterView

__all__ = ['LCviz']


@unit_converter('custom-lcviz')
class UnitConverter:
    def equivalent_units(self, data, cid, units):
        return set(list(map(str, u.Unit(units).find_equivalent_units(include_prefix_units=True))))

    def to_unit(self, data, cid, values, original_units, target_units):
        # for some reason, glue is trying to request a change for cid='flux' from d to electron / s
        if target_units not in self.equivalent_units(data, cid, original_units):
            return values
        return (values * u.Unit(original_units)).to_value(u.Unit(target_units))


glue_settings.UNIT_CONVERTER = 'custom-lcviz'

custom_components = {'plugin-ephemeris-select': 'components/plugin_ephemeris_select.vue'}

# Register pure vue component. This allows us to do recursive component instantiation only in the
# vue component file
for name, path in custom_components.items():
    ipyvue.register_component_from_file(None, name,
                                        os.path.join(os.path.dirname(__file__), path))


def _get_range_subset_bounds(self, subset_state, *args, **kwargs):
    viewer = self._jdaviz_helper.default_time_viewer._obj
    light_curve = viewer.data()[0]
    reference_time = light_curve.meta['reference_time']
    if viewer:
        # TODO: use display units once implemented in Glue for ScatterViewer
        # units = u.Unit(viewer.state.x_display_unit)
        units = u.Unit(viewer.time_unit)
    else:  # pragma: no cover
        raise ValueError("Unable to find time axis units")

    region = reference_time + u.Quantity([subset_state.lo * units, subset_state.hi * units])

    return [{"name": subset_state.__class__.__name__,
             "glue_state": subset_state.__class__.__name__,
             "region": region,
             "subset_state": subset_state}]


def _link_new_data(app, reference_data=None, data_to_be_linked=None):
    dc = app.data_collection
    ref_data = dc[reference_data] if reference_data else dc[0]
    linked_data = dc[data_to_be_linked] if data_to_be_linked else dc[-1]

    new_link = LinkSame(
        cid1=ref_data.world_component_ids[0],
        cid2=linked_data.world_component_ids[0],
        data1=ref_data,
        data2=linked_data,
        labels1=ref_data.label,
        labels2=linked_data.label
    )

    dc.add_link(new_link)
    return


def _get_display_unit(app, axis):
    if app._jdaviz_helper is None or app._jdaviz_helper.plugins.get('Unit Conversion') is None:  # noqa
        # fallback on native units (unit conversion is not enabled)
        if axis == 'time':
            return u.d
        elif axis == 'flux':
            return app._jdaviz_helper.default_time_viewer._obj.data()[0].flux.unit
        else:
            raise ValueError(f"could not find units for axis='{axis}'")
    try:
        # TODO: need to implement and add unit conversion plugin for this to be able to work
        return getattr(app._jdaviz_helper.plugins.get('Unit Conversion')._obj,
                       f'{axis}_unit_selected')
    except AttributeError:
        raise ValueError(f"could not find display unit for axis='{axis}'")


class LCviz(ConfigHelper):
    _default_configuration = {
        'settings': {'configuration': 'lcviz',
                     'visible': {'menu_bar': False,
                                 'toolbar': True,
                                 'tray': True,
                                 'tab_headers': True},
                     'dense_toolbar': False,
                     'context': {'notebook': {'max_height': '600px'}}},
        'toolbar': ['g-data-tools', 'g-subset-tools', 'g-viewer-creator', 'lcviz-coords-info'],
        'tray': ['lcviz-metadata-viewer', 'flux-column',
                 'lcviz-plot-options', 'lcviz-subset-tools',
                 'lcviz-markers', 'time-selector',
                 'stitch', 'flatten', 'frequency-analysis', 'ephemeris',
                 'binning', 'lcviz-export'],
        'viewer_area': [{'container': 'col',
                         'children': [{'container': 'row',
                                       'viewers': [{'name': 'flux-vs-time',
                                                    'plot': 'lcviz-time-viewer',
                                                    'reference': 'flux-vs-time'}]}]}]}

    _component_ids = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # override jdaviz behavior to support temporal subsets
        self.app._get_range_subset_bounds = (
            lambda *args, **kwargs: _get_range_subset_bounds(self.app, *args, **kwargs)
        )

        self.app._link_new_data = (
            lambda *args, **kwargs: _link_new_data(self.app, *args, **kwargs)
        )

        self.app._get_display_unit = (
            lambda *args, **kwargs: _get_display_unit(self.app, *args, **kwargs)
        )

        # inject custom css from lcviz_style.vue (on top of jdaviz styles)
        self.app._add_style((__file__, 'lcviz_style.vue'))

        # set the link to read the docs
        self.app.vdocs = 'latest' if 'dev' in __version__ else 'v'+__version__
        self.app.docs_link = f"https://lcviz.readthedocs.io/en/{self.app.vdocs}"
        for plugin in self.plugins.values():
            # NOTE that plugins that need to override upstream docs_link should do so in
            # an @observe('vdocs') rather than the init, since plugin instances have
            # already been initialized
            plugin._obj.vdocs = self.app.vdocs

    def load_data(self, data, data_label=None, extname=None):
        """
        Load data into LCviz.

        Parameters
        ----------
        data : obj or str
            File name or object to be loaded. Supported formats include:

            * ``'filename.fits'`` (or any extension that ``astropy.io.fits``
              supports)
            * `~lightkurve.LightCurve` (extracts the default flux column)
        data_label : str or `None`
            Data label to go with the given data. If not given, this is
            automatically determined from filename or randomly generated.
            The final label shown in LCviz may have additional information
            appended for clarity.
        extname : str or `None`
            Used for DVT parsing if only a single TCE from a multi-TCE file should be
            loaded. Formatted as 'TCE_1', 'TCE_2', etc.
        """
        # Determine if we're loading a DVT file, which has a separate parser
        if isinstance(data, str):
            header = getheader(data)
            if (header['TELESCOP'] == 'TESS' and 'CREATOR' in header and
                    'DvTimeSeriesExporter' in header['CREATOR']):
                super().load_data(data=data,
                                  parser_reference='tess_dvt_parser',
                                  data_label=data_label,
                                  extname=extname)
                return

        super().load_data(
            data=data,
            parser_reference='light_curve_parser',
            data_label=data_label
        )

    def get_data(self, data_label=None, cls=LightCurve, subset=None):
        """
        Returns data with name equal to data_label of type cls with subsets applied from
        subset_to_apply.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        cls : light curve class, optional
            The type that data will be returned as.
        subset : str, optional
            Subset that is to be applied (as a mask) to the data before it is returned.

        Returns
        -------
        data : cls
            Data is returned as type cls with subsets applied.
        """
        return super()._get_data(data_label=data_label, mask_subset=subset, cls=cls)

    @property
    def default_time_viewer(self):
        tvs = [viewer for vid, viewer in self.app._viewer_store.items()
               if isinstance(viewer, TimeScatterView)]
        if not len(tvs):
            raise ValueError("no time viewers exist")
        return tvs[0].user_api

    @property
    def _has_cube_data(self):
        for data in self.app.data_collection:
            if data.ndim == 3:
                return True
        return False

    @property
    def _tray_tools(self):
        """
        Access API objects for plugins in the app toolbar.

        Returns
        -------
        plugins : dict
            dict of plugin objects
        """
        # TODO: provide user-friendly labels, user API, and move upstream to be public
        # for now this is just useful for dev-debugging access to toolbar entries
        from ipywidgets.widgets import widget_serialization
        return {item['name']: widget_serialization['from_json'](item['widget'], None)
                for item in self.app.state.tool_items}

    def _get_clone_viewer_reference(self, reference):
        base_name = reference.split("[")[0]
        name = base_name
        ind = 0
        while name in self.viewers.keys():
            ind += 1
            name = f"{base_name}[{ind}]"
        return name

    def _phase_comp_lbl(self, component):
        return f'phase:{component}'

    def _set_data_component(self, data, component_label, values):
        if component_label in self._component_ids:
            component_id = self._component_ids[component_label]
        else:
            existing_components = [component.label for component in data.components]
            if component_label in existing_components:
                component_id = data.components[existing_components.index(component_label)]
            else:
                component_id = ComponentID(component_label)
                self._component_ids[component_label] = component_id

        if component_id in data.components:
            data.update_components({component_id: values})
        else:
            data.add_component(values, component_id)



import os

from astropy.io import fits
from astropy.table import Table
from glue.config import data_translator
from jdaviz.core.registries import data_parser_registry
import lightkurve
import numpy as np

from lcviz.viewers import PhaseScatterView, TimeScatterView
from lcviz.plugins.plot_options import PlotOptions

__all__ = ["light_curve_parser"]

mission_sub_intervals = {
    'kepler': {'prefix': 'Q', 'card': 'QUARTER'},
    'k2': {'prefix': 'C', 'card': 'CAMPAIGN'},
    'tess': {'prefix': 'S', 'card': 'SECTOR'},
    'tess dvt': {'prefix': '', 'card': 'EXTNAME'}
}


@data_parser_registry("tess_dvt_parser")
def tess_dvt_parser(app, file_obj, data_label=None, show_in_viewer=True, **kwargs):
    '''
    Read a TESS DVT file and create a lightkurve object
    '''
    hdulist = fits.open(file_obj)
    ephem_plugin = app._jdaviz_helper.plugins.get('Ephemeris', None)
    extname = kwargs.pop('extname')

    # Loop through the TCEs in the file. If we only want one (specified by
    # `extname` keyword) then only load that one into the viewers and ephemeris.
    for hdu in hdulist[1:]:
        data = Table(hdu.data)
        # don't load some columns with names that may
        # conflict with components generated later by lcviz
        data.remove_column('PHASE')
        data.remove_column('CADENCENO')
        # Remove rows that have NaN data
        data = data[~np.isnan(data['LC_INIT'])]
        header = hdu.header
        time_offset = int(header['TUNIT1'] .split('- ')[1].split(',')[0])
        data['TIME'] += time_offset
        lc = lightkurve.LightCurve(data=data,
                                   time=data['TIME'],
                                   flux=data['LC_INIT'],
                                   flux_err=data['LC_INIT_ERR'])
        lc.meta = hdulist[0].header
        lc.meta['MISSION'] = 'TESS DVT'
        lc.meta['FLUX_ORIGIN'] = "LC_INIT"
        lc.meta['EXTNAME'] = header['EXTNAME']

        if extname is not None and header['EXTNAME'] != extname:
            show_ext_in_viewer = False
        else:
            show_ext_in_viewer = show_in_viewer

        light_curve_parser(app, lc, data_label=data_label,
                           show_in_viewer=show_ext_in_viewer, **kwargs)

        # add ephemeris information from the DVT extension
        if ephem_plugin is not None and show_ext_in_viewer:
            ephem_plugin.period = header['TPERIOD']
            ephem_plugin.t0 = header['TEPOCH'] + time_offset - app.data_collection[0].coords.reference_time.jd  # noqa


@data_parser_registry("light_curve_parser")
def light_curve_parser(app, file_obj, data_label=None, show_in_viewer=True, **kwargs):
    # load a LightCurve or TargetPixelFile object:
    cls_with_translator = (
        lightkurve.LightCurve,
        lightkurve.targetpixelfile.KeplerTargetPixelFile,
        lightkurve.targetpixelfile.TessTargetPixelFile
    )

    # load local FITS file from disk by its path:
    if isinstance(file_obj, str) and os.path.exists(file_obj):
        if data_label is None:
            data_label = os.path.splitext(os.path.basename(file_obj))[0]

        light_curve = lightkurve.read(file_obj)

    elif isinstance(file_obj, cls_with_translator):
        light_curve = file_obj
    else:
        raise NotImplementedError(f"could not parse light_curve with type {type(file_obj)}")

    # handle flux_origin default
    mission = light_curve.meta.get('MISSION', '').lower()
    flux_origin = light_curve.meta.get('FLUX_ORIGIN', None)  # i.e. PDCSAP or SAP

    # make a data label:
    if data_label is None:
        data_label = light_curve.meta.get('OBJECT', 'Light curve')

        if isinstance(light_curve, lightkurve.targetpixelfile.TargetPixelFile):
            data_label += '[TPF]'
        elif mission in mission_sub_intervals:
            # the sub-interval label is something like "Q9" for Kepler or
            # "S9" for TESS. If it's already in the proposed data label, skip;
            # otherwise, append it.
            sub_interval_label = (
                f' [{mission_sub_intervals[mission]["prefix"]}'
                f'{light_curve.meta.get(mission_sub_intervals[mission]["card"])}]'
            )
            data_label += sub_interval_label

    if flux_origin == 'flux' or (flux_origin is None and 'flux' in getattr(light_curve, 'columns', [])):  # noqa
        # then make a copy of this column so it won't be lost when changing with the flux_column
        # plugin
        light_curve['flux:orig'] = light_curve['flux']
        if 'flux_err' in light_curve.columns:
            light_curve['flux:orig_err'] = light_curve['flux_err']
        light_curve.meta['FLUX_ORIGIN'] = 'flux:orig'

    if 'FILENAME' in light_curve.meta:
        light_curve.meta['FILENAME'] = os.path.basename(light_curve.meta['FILENAME'])

    data = _data_with_reftime(app, light_curve)
    app.add_data(data, data_label)

    if isinstance(light_curve, lightkurve.targetpixelfile.TargetPixelFile):
        # ensure an image/cube/TPF viewer exists
        # TODO: move this to an event listener on add_data so that we can also remove when empty?
        from jdaviz.core.events import NewViewerMessage
        from lcviz.viewers import CubeView
        if show_in_viewer:
            found_viewer = False
            for viewer_id, viewer in app._viewer_store.items():
                if isinstance(viewer, CubeView):
                    app.add_data_to_viewer(viewer_id, data_label)
                    found_viewer = True
            if not found_viewer:
                app._on_new_viewer(NewViewerMessage(CubeView, data=None, sender=app),
                                   vid='image', name='image')
                app.add_data_to_viewer('image', data_label)

        # set TPF viewer's stretch to custom defaults:
        plot_options_plugin = PlotOptions(app=app)
        if plot_options_plugin is not None:
            plot_options_plugin._default_tpf_stretch()

    else:
        if show_in_viewer:
            for viewer_id, viewer in app._viewer_store.items():
                if isinstance(viewer, (TimeScatterView, PhaseScatterView)):
                    app.add_data_to_viewer(viewer_id, data_label)

            # add to any known phase viewers
            ephem_plugin = app._jdaviz_helper.plugins.get('Ephemeris', None)
            if ephem_plugin is not None:
                for viewer in ephem_plugin._obj._get_phase_viewers():
                    app.add_data_to_viewer(viewer.reference, data_label)


def _data_with_reftime(app, light_curve):
    # grab the first-found reference time in the data collection:
    ff_reference_time = None
    for existing_data in app.data_collection:
        if hasattr(existing_data, 'meta') and 'reference_time' in existing_data.meta:
            ff_reference_time = existing_data.meta.get('reference_time', None)
            if ff_reference_time is not None:
                break

    # convert to glue Data manually, so we may edit the `dt` component if necessary:
    handler, _ = data_translator.get_handler_for(light_curve)
    return handler.to_data(light_curve, reference_time=ff_reference_time)



def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load_data(lc)
    # lcviz.show()

    metadata = lcviz.plugins['Metadata']
    print(f"dataset choices: {metadata.dataset.choices}")
    metadata.dataset = metadata.dataset.choices[0]
    print(metadata.meta)



import pytest


@pytest.mark.remote_data
def test_tray_viewer_creator(helper, light_curve_like_kepler_quarter):
    # additional coverage in test_plugin_ephemeris
    helper.load_data(light_curve_like_kepler_quarter)
    vc = helper._tray_tools['g-viewer-creator']

    assert len(helper.viewers) == 1
    assert len(vc.viewer_types) == 2  # time and default phase
    vc.vue_create_viewer('flux-vs-time')
    assert len(helper.viewers) == 2

    # TODO: replace with test fixture
    from lightkurve import search_targetpixelfile
    tpf = search_targetpixelfile("KIC 001429092",
                                 mission="Kepler",
                                 cadence="long",
                                 quarter=10).download()
    helper.load_data(tpf)
    assert len(helper.viewers) == 3  # image viewer added by default

    assert len(vc.viewer_types) == 3  # time, default phase, cube
    vc.vue_create_viewer('image')
    assert len(helper.viewers) == 4



from astropy.io.fits.card import Undefined
from traitlets import observe

from jdaviz.configs.default.plugins import MetadataViewer
from jdaviz.core.registries import tray_registry

__all__ = ['MetadataViewer']


# monkeypatch astropy.io.fits.card.Undefined to show an empty string
# instead of '<astropy.io.fits.card.Undefined object at 0x29f5b94d0>'
Undefined.__str__ = lambda x: ''


@tray_registry('lcviz-metadata-viewer', label="Metadata")
class MetadataViewer(MetadataViewer):
    """
    See the :ref:`Metadata Viewer Plugin Documentation <metadata-viewer>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to expose the metadata.
    * :attr:`show_primary` : bool
      Whether to show MEF primary header metadata instead.
    * :attr:`metadata`:
      Read-only metadata. If the data is loaded from a multi-extension FITS file,
      this can be the extension header or the primary header, depending on
      ``show_primary`` setting.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @observe('vdocs')
    def _update_docs_link(self, *args):
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#metadata-viewer"  # noqa


# Licensed under a 3-clause BSD style license - see LICENSE.rst

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

from . import utils  # noqa
from .utils import enable_hot_reloading  # noqa: F401
from . import state  # noqa

# Expose subpackage API at package level.
from .plugins import *  # noqa
from .parsers import *  # noqa
from .tools import *  # noqa
from .viewers import *  # noqa
from .helper import *  # noqa
